use {
    b64::{Decoder, Encoder},
    getopts::Options,
    std::{
        env,
        error::Error,
        fs::File,
        io::{self, BufReader, BufWriter, Write},
        os::fd::AsFd,
        path::Path,
    },
};

#[derive(PartialEq)]
enum Operation {
    Encode,
    Decode,
}

enum Input {
    Stdin,
    Filename(String),
}

fn print_usage(program: &str, opts: Options) {
    let brief = format!("Usage: {} FILE [options]", program);
    print!("{}", opts.usage(&brief));
}

fn main() -> Result<(), Box<dyn Error>> {
    let args = env::args().collect::<Vec<String>>();
    let progname = Path::new(&args[0])
        .file_name()
        .map(|x| x.to_string_lossy())
        .unwrap();
    let mut opts = Options::new();
    opts.optflag("h", "help", "print this help message");
    opts.optflag("d", "decode", "decode instead of encoding");
    opts.optflag("i", "ignore", "ignore whitespace when decoding");
    opts.optopt("o", "output", "output to FILE instead of STDOUT", "FILE");
    opts.optopt("w", "wrap", "wrap at columns", "COLS");
    let matches = match opts.parse(&args[1..]) {
        Ok(m) => m,
        Err(f) => {
            panic!("{}", f.to_string())
        }
    };
    if matches.opt_present("h") {
        print_usage(&progname, opts);
        return Ok(());
    }
    let op = if matches.opt_present("d") {
        Operation::Decode
    } else {
        Operation::Encode
    };
    let ignore = matches.opt_present("i");
    let wrap: usize = if let Some(w) = matches.opt_str("w") {
        w.parse()?
    } else {
        76
    };

    let mut infiles = matches
        .free
        .iter()
        .map(|x| {
            if x == "-" {
                Input::Stdin
            } else {
                Input::Filename(x.to_string())
            }
        })
        .collect::<Vec<_>>();
    if infiles.is_empty() {
        infiles.push(Input::Stdin)
    }
    for f in &infiles {
        let writer = if let Some(o) = matches.opt_str("o") {
            if matches.free.len() < 2 || op == Operation::Encode {
                BufWriter::new(File::open(o)?)
            } else {
                panic!("Attempt to decode multiple files into single file");
            }
        } else {
            BufWriter::new(File::from(io::stdout().as_fd().try_clone_to_owned()?))
        };
        let reader = match f {
            Input::Stdin => BufReader::new(File::from(io::stdin().as_fd().try_clone_to_owned()?)),
            Input::Filename(f) => BufReader::new(File::open(f)?),
        };
        match op {
            Operation::Decode => {
                let decoder = Decoder::new(reader, writer, None, ignore);
                decoder.decode()?;
            }
            Operation::Encode => {
                let mut encoder = Encoder::new(reader, writer, None, Some(wrap));
                encoder.encode()?;
                encoder.output().write_all(&[b'\n'])?;
            }
        }
    }

    Ok(())
}
